local BACKWARD_COMPATIBILITY = 13000

local FLASH_HIGH = 0.8
local FLASH_VALUES = { FLASH_HIGH }
local ONE_VALUES = { 1 }
local FLASH_DURATION = 0.4
local ENERGY_INC_DURATION = 0.1

local DURATIONS = {
	COMBO_SHOW_DURATION = { 0.1, 0.1 },
	COMBO_HIDE_DURATION = { 0.2, 0.2 },
	MAINFRAME_SHOW_DURATION = { 0.1, 0.1 },
	MAINFRAME_HIDE_DURATION = { 0.2, 0.2 },
	ENERGY_DEC_DURATION = { 0.1, 0.1 },
}
local INSTANT_DURATION = 0.01

local ROGUE = 1
local DRUID = 2

local ENERGY_SUBBAR_NUM = 5

local SPELL_ID_PROWL = 5215
local SPELL_NAME_PROWL

ComergyBarTextures = {
	{ "Patt", "Interface\\AddOns\\Comergy\\textures\\Patt" },
	{ "Flat", "Interface\\AddOns\\Comergy\\textures\\Flat" },
	{ "Smth", "Interface\\AddOns\\Comergy\\textures\\Smth" },
	{ "Alum", "Interface\\AddOns\\Comergy\\textures\\Alum" },
	{ "3D", "Interface\\AddOns\\Comergy\\textures\\3d" },
	{ "Pat2", "Interface\\AddOns\\Comergy\\textures\\Pat2" },
}

ComergyTextFonts = {
	{ "Chat", "ChatFontNormal" },
	{ "Combat", "NumberFontNormal" },
	{ "System", "GameFontNormal" },
}

local status = {
	initialized = false,

	playerClass,
	playerInCombat,
	playerInStealth,
	playerEnergyEnabled,

	maxEnergy,
	curEnergy,
	energyFlashing,

	curCombo,
	comboFlashing,
}
local playerGUID

local comboBars = {}
local energyBars = {}

-- LOCAL FUNCTIONS --

local InitObject, GradientObject, UpdateObject, SetStatusBarValue, MainFrameShow, MainFrameToggle, SplitEnergyBar
local EnergyChanged, MaxEnergyChanged, FrameResize, ComboChanged, Initialize, ReEnterWorld
local EventHandlers, RestorePosition, ResetObject, TextChanged, TextStyleChanged, BGResize

function InitObject(object, initValues)
	object.curValue = {}
	object.endValue = {}
	object.deltaValue = {}
	for i = 1, #(initValues) do
		object.curValue[i] = initValues[i]
		object.endValue[i] = initValues[i]
		object.deltaValue[i] = 0
	end
end

function ResetObject(object, resetValues)
	for i = 1, #(resetValues) do
		object.curValue[i] = resetValues[i]
		object.endValue[i] = resetValues[i]
		object.deltaValue[i] = 0
	end
end

function GradientObject(object, idx, duration, endValue)
	if (object.endValue[idx] ~= endValue) then
		object.deltaValue[idx] = (endValue - object.curValue[idx]) / duration
		object.endValue[idx] = endValue
	end
end

function UpdateObject(object, elapsed)
	for i = 1, #(object.deltaValue) do
		if (object.deltaValue[i] ~= 0) then
			local curValue = object.curValue[i] + object.deltaValue[i] * elapsed
			if ((curValue - object.endValue[i]) * object.deltaValue[i] >= 0) then
				object.curValue[i] = object.endValue[i]
				object.deltaValue[i] = 0
			else
				object.curValue[i] = curValue
			end
		end
	end
end

function SetStatusBarValue(bar, value)
	if (value == 0) then
		bar:GetStatusBarTexture():Hide()
		return
	else
		bar:GetStatusBarTexture():Show()
	end

	value = value / (bar.max - bar.min)

	if (bar.direction < 3) then
		bar:GetStatusBarTexture():SetWidth(bar.len * value);
	else
		bar:GetStatusBarTexture():SetHeight(bar.len * value);
	end

	if (bar.direction == 1) then
		bar:GetStatusBarTexture():SetTexCoord(0, value, 0, 1);
	elseif (bar.direction == 2) then
		bar:GetStatusBarTexture():SetTexCoord(value, 0, 0, 1);
	elseif (bar.direction == 3) then
		bar:GetStatusBarTexture():SetTexCoord(value, 0, 0, 0, value, 1, 0, 1);
	elseif (bar.direction == 4) then
		bar:GetStatusBarTexture():SetTexCoord(0, 1, value, 1, 0, 0, value, 0);
	end
end

function MainFrameShow(show)
	if (show) then
		GradientObject(ComergyMainFrame, 1, DURATIONS["MAINFRAME_SHOW_DURATION"][1], 1)
		ComergyMainFrame:Show()
	else
		GradientObject(ComergyMainFrame, 1, DURATIONS["MAINFRAME_HIDE_DURATION"][1], 0)
	end
end

function MainFrameToggle()
	if (not status.playerEnergyEnabled) then
		MainFrameShow(false)
	elseif ((status.curCombo ~= 0) or (status.playerInCombat)) then
		MainFrameShow(true)
	else
		if (not Comergy_Config.ShowInCombat) then
			MainFrameShow(true)
		else
			if (((not Comergy_Config.ShowWhenEnergyNotFull) or (status.curEnergy == status.maxEnergy)) and
				((not Comergy_Config.ShowInStealth) or (not status.playerInStealth))) then
				MainFrameShow(false)
			else
				MainFrameShow(true)
			end
		end
	end
end

function SplitEnergyBar()
	energyBars = {}

	local n = 1
	local thresholds = {}
	energyBars[1] = getglobal("ComergyEnergyBar1")
	energyBars[1].min = 0
	energyBars[1].minColor = Comergy_Config.EnergyColor0
	for i = 1, ENERGY_SUBBAR_NUM - 1 do
		if (Comergy_Config["SplitEnergy"..i]) then
			local th = Comergy_Config["EnergyThreshold"..i]
			if ( th < status.maxEnergy) then
				local temp = { th, i }
				tinsert(thresholds, temp)
			end
		end
	end
	table.sort(thresholds, function(a, b) return a[1] < b[1] end)

	local lastThreshold = 0
	for i = 1, #(thresholds) do
		if (thresholds[i][1] ~= lastThreshold) then
			n = n + 1
			energyBars[n] = getglobal("ComergyEnergyBar"..n)
			energyBars[n - 1].max = thresholds[i][1]
			energyBars[n].min = thresholds[i][1]
			energyBars[n - 1].maxColor = Comergy_Config["EnergyColor"..thresholds[i][2]]
			energyBars[n].minColor = Comergy_Config["EnergyColor"..thresholds[i][2]]
			lastThreshold = thresholds[i][1]
		end
	end

	energyBars[n].max = status.maxEnergy
	energyBars[n].maxColor = Comergy_Config["EnergyColor"..ENERGY_SUBBAR_NUM]

	local lenPerEnergy = (Comergy_Config.Width - Comergy_Config.Spacing * (n - 1)) / status.maxEnergy
	local anchorPointV = "TOP"
	local relAnchorPointV = "BOTTOM"
	local anchorPointH = "RIGHT"
	local relAnchorPointH = "LEFT"
	local direction

	if (Comergy_Config.FlipBars) then
		if (Comergy_Config.VerticalBars) then
			anchorPointH, relAnchorPointH = relAnchorPointH, anchorPointH
		else
			anchorPointV, relAnchorPointV = relAnchorPointV, anchorPointV
		end
	end
	if (Comergy_Config.FlipOrientation) then
		if (Comergy_Config.VerticalBars) then
			anchorPointV, relAnchorPointV = relAnchorPointV, anchorPointV				
		else
			anchorPointH, relAnchorPointH = relAnchorPointH, anchorPointH
		end
		direction = 2
	else
		direction = 1
	end

	if (Comergy_Config.VerticalBars) then
		direction = direction + 2
	end

	for i = 1, n do
		local left = energyBars[i].min * lenPerEnergy + (i - 1) * Comergy_Config.Spacing
		local right = energyBars[i].max * lenPerEnergy + (i - 1) * Comergy_Config.Spacing
		local top = Comergy_Config.EnergyHeight
		local bottom = 0
		energyBars[i].len = right - left

		if (Comergy_Config.FlipBars) then
			top = -top
		end
		if (Comergy_Config.FlipOrientation) then
			left, right = -left, -right
		end

		energyBars[i].direction = direction
		energyBars[i]:SetStatusBarTexture(ComergyBarTextures[Comergy_Config.BarTexture][2])

		if (Comergy_Config.VerticalBars) then
			left, bottom = bottom, left
			right, top = top, right
			energyBars[i]:GetStatusBarTexture():ClearAllPoints()
			energyBars[i]:GetStatusBarTexture():SetPoint(relAnchorPointV, 0, 0)
			energyBars[i]:GetStatusBarTexture():SetWidth(Comergy_Config.EnergyHeight)
		else
			energyBars[i]:GetStatusBarTexture():ClearAllPoints()
			energyBars[i]:GetStatusBarTexture():SetPoint(relAnchorPointH, 0, 0)
			energyBars[i]:GetStatusBarTexture():SetHeight(Comergy_Config.EnergyHeight)
		end

		energyBars[i]:ClearAllPoints()
		energyBars[i]:SetPoint(relAnchorPointV .. relAnchorPointH, left, bottom)
		energyBars[i]:SetPoint(anchorPointV .. anchorPointH, energyBars[i]:GetParent(), relAnchorPointV .. relAnchorPointH, right, top)

		if ((top == 0) or (right == 0)) then
			energyBars[i]:Hide()
		else
			energyBars[i]:Show()
		end
	end

	for i = n + 1, ENERGY_SUBBAR_NUM do
		getglobal("ComergyEnergyBar"..i):Hide()
	end
end

function EnergyChanged(inc)
	local energyDuration
	if (inc) then
		energyDuration = ENERGY_INC_DURATION
		for i = 1, ENERGY_SUBBAR_NUM - 1 do
			if (status.curEnergy == Comergy_Config["EnergyThreshold"..i]) then
				if (((not Comergy_Config.ShowInCombat) or (status.playerInCombat)) and (Comergy_Config["SoundEnergy"..i])) then
					PlaySoundFile("Interface\\AddOns\\Comergy\\sound\\energytick.mp3")
				end
			end
		end
		if (((not Comergy_Config.ShowInCombat) or (status.playerInCombat)) and (Comergy_Config["SoundEnergy"..ENERGY_SUBBAR_NUM]) and (status.curEnergy == status.maxEnergy)) then
			PlaySoundFile("Interface\\AddOns\\Comergy\\sound\\energytick.mp3")
		end
	else
		energyDuration = DURATIONS["ENERGY_DEC_DURATION"][1]
	end
	for i = 1, #(energyBars) do
		if (energyBars[i].min > status.curEnergy) then
			if (energyBars[i].endValue[1] ~= 0) then
				GradientObject(energyBars[i], 1, INSTANT_DURATION, 0)
				if (not Comergy_Config.UnifiedEnergyColor) then
					local j
					for j = 1, 3 do
						GradientObject(energyBars[i], j + 1, energyDuration, energyBars[i].minColor[j])
					end
				end
			end
		elseif (energyBars[i].max < status.curEnergy) then
			if (energyBars[i].endValue[1] ~= energyBars[i].max - energyBars[i].min) then
				GradientObject(energyBars[i], 1, energyDuration, energyBars[i].max - energyBars[i].min)
				if (not Comergy_Config.UnifiedEnergyColor) then
					local j
					for j = 1, 3 do
						GradientObject(energyBars[i], j + 1, energyDuration, energyBars[i].maxColor[j])
					end
				end
			end
		else
			energyBars[i]:GetStatusBarTexture():Show()
			GradientObject(energyBars[i], 1, energyDuration, status.curEnergy - energyBars[i].min)
			local j
			for j = 1, 3 do
				local color = energyBars[i].minColor[j] + (energyBars[i].maxColor[j] - energyBars[i].minColor[j]) / (energyBars[i].max - energyBars[i].min) * (status.curEnergy - energyBars[i].min)
				if (Comergy_Config.UnifiedEnergyColor) then
					local k
					for k = 1, #(energyBars) do
						GradientObject(energyBars[k], j + 1, energyDuration, color)
					end
				else
					GradientObject(energyBars[i], j + 1, energyDuration, color)
				end
			end
		end
	end

	TextChanged()

	if (status.curEnergy == status.maxEnergy) then
		MainFrameToggle()
	end
end

function MaxEnergyChanged()
	status.maxEnergy = UnitManaMax("player")
	SplitEnergyBar()
	TextChanged()
	MainFrameToggle()
end	

function ComboChanged()
	local combo = GetComboPoints("player", "target")

	if ((Comergy_Config.ComboFlash) and (combo == 5)) then
		if (status.comboFlashing == 0) then
			status.comboFlashing = 2
			for i = 1, #(comboBars) do
				ResetObject(comboBars[i], FLASH_VALUES)
				comboBars[i]:SetAlpha(FLASH_HIGH)
				GradientObject(comboBars[i], 1, FLASH_DURATION, Comergy_Config.ComboBGAlpha)
				if ((Comergy_Config.UnifiedComboColor) or (i == 5)) then
					local color = Comergy_Config["ComboColor5"]
					local j
					for j = 1, 3 do
						GradientObject(comboBars[i], j + 1, DURATIONS["COMBO_SHOW_DURATION"][1], color[j])
					end
				end
			end
		end
	else
		status.comboFlashing = 0

		for i = 1, combo do
			GradientObject(comboBars[i], 1, DURATIONS["COMBO_SHOW_DURATION"][1], 1)
			if (Comergy_Config.UnifiedComboColor) then
				local color = Comergy_Config["ComboColor"..combo]
				local j
				for j = 1, 3 do
					GradientObject(comboBars[i], j + 1, DURATIONS["COMBO_SHOW_DURATION"][1], color[j])
				end
			end
		end

		for i = combo + 1, 5 do
			GradientObject(comboBars[i], 1, DURATIONS["COMBO_HIDE_DURATION"][1], Comergy_Config.ComboBGAlpha)
			if (Comergy_Config.UnifiedComboColor) then
				local j
				for j = 1, 3 do
					GradientObject(comboBars[i], j + 1, DURATIONS["COMBO_HIDE_DURATION"][1], Comergy_Config.ComboColor0[j])
				end
			end
		end
	end
	
	if ((combo ~= status.curCombo) and (combo ~= 0)) then
		if (Comergy_Config["SoundCombo"..combo]) then
			PlaySoundFile("Interface\\AddOns\\Comergy\\sound\\combo"..combo..".mp3")
		end
	end

	status.curCombo = combo
	TextChanged()
	MainFrameToggle()
end

function FrameResize()
	local w = Comergy_Config.Width
	local space
	if (Comergy_Config.ComboHeight * Comergy_Config.EnergyHeight ~= 0) then
		space = Comergy_Config.Spacing
	else
		space = 0
	end
	local h = Comergy_Config.ComboHeight + Comergy_Config.EnergyHeight + space
	if (Comergy_Config.VerticalBars) then
		w, h = h, w
	end
	ComergyMainFrame:SetWidth(w)
	ComergyMainFrame:SetHeight(h)

	local comboLength = (Comergy_Config.Width - 4 * Comergy_Config.Spacing) / 5

	local anchorPointV = "BOTTOM"
	local relAnchorPointV = "TOP"
	local anchorPointH = "RIGHT"
	local relAnchorPointH = "LEFT"

	if (Comergy_Config.FlipBars) then
		if (Comergy_Config.VerticalBars) then
			anchorPointH, relAnchorPointH = relAnchorPointH, anchorPointH
		else
			anchorPointV, relAnchorPointV = relAnchorPointV, anchorPointV
		end
	end
	if (Comergy_Config.FlipOrientation) then
		if (Comergy_Config.VerticalBars) then
			anchorPointV, relAnchorPointV = relAnchorPointV, anchorPointV
		else
			anchorPointH, relAnchorPointH = relAnchorPointH, anchorPointH
		end
	end
	if (Comergy_Config.VerticalBars) then
		anchorPointV, relAnchorPointV = relAnchorPointV, anchorPointV
		anchorPointH, relAnchorPointH = relAnchorPointH, anchorPointH
	end

	local i
	for i = 1, 5 do
		comboBars[i]:SetStatusBarTexture(ComergyBarTextures[Comergy_Config.BarTexture][2])
		local left = (i - 1) * (Comergy_Config.Spacing + comboLength)
		local right = left + comboLength
		local top = -Comergy_Config.ComboHeight
		local bottom = 0

		if (Comergy_Config.FlipBars) then
			top = -top
		end
		if (Comergy_Config.FlipOrientation) then
			left, right = -left, -right
		end

		if (Comergy_Config.VerticalBars) then
			left, bottom = bottom, left
			right, top = top, right
			comboBars[i]:GetStatusBarTexture():SetTexCoord(1, 0, 0, 0, 1, 1, 0, 1);
		else
			comboBars[i]:GetStatusBarTexture():SetTexCoord(0, 1, 0, 1);
		end

		comboBars[i]:ClearAllPoints()
		comboBars[i]:SetPoint(relAnchorPointV .. relAnchorPointH, left, bottom)
		comboBars[i]:SetPoint(anchorPointV .. anchorPointH, comboBars[i]:GetParent(), relAnchorPointV .. relAnchorPointH, right, top)
	end

	ComergyEnergyText:ClearAllPoints()
	ComergyComboText:ClearAllPoints()
	if (Comergy_Config.VerticalBars) then
		ComergyEnergyText:SetPoint("BOTTOM", ComergyMainFrame, "TOP", 0, 3)
		ComergyComboText:SetPoint("TOP", ComergyMainFrame, "BOTTOM", 0, -3)
	else
		ComergyEnergyText:SetPoint("RIGHT", ComergyMainFrame, "LEFT", -3, 0)
		ComergyComboText:SetPoint("LEFT", ComergyMainFrame, "RIGHT", 3, 0)
	end
end

function BGResize()
	ComergyMovingFrame:ClearAllPoints()

	local left, bottom = -Comergy_Config.Spacing, -Comergy_Config.Spacing
	local right, top = Comergy_Config.Spacing, Comergy_Config.Spacing

	if (not Comergy_Config.TextCenter) then
		if (Comergy_Config.VerticalBars) then
			local diff = (ComergyEnergyText:GetWidth() - ComergyMainFrame:GetWidth()) / 2
			diff = (diff > 0) and diff or 0
			if (Comergy_Config.EnergyText) then
				top = Comergy_Config.Spacing + ComergyEnergyText:GetHeight()
				left = left - diff
				right = right + diff
				diff = 0
			end
			if (Comergy_Config.ComboText) then
				bottom = -(Comergy_Config.Spacing + ComergyEnergyText:GetHeight())
				left = left - diff
				right = right + diff
			end
		else
			local diff = (ComergyEnergyText:GetHeight() - ComergyMainFrame:GetHeight()) / 2
			diff = (diff > 0) and diff or 0
			if (Comergy_Config.EnergyText) then
				left = -(Comergy_Config.Spacing + ComergyEnergyText:GetWidth())
				top = top + diff
				bottom = bottom - diff
				diff = 0
			end
			if (Comergy_Config.ComboText) then
				right = Comergy_Config.Spacing + ComergyEnergyText:GetWidth()
				top = top + diff
				bottom = bottom - diff
			end
		end
	end

	ComergyMovingFrame:SetPoint("TOPLEFT", ComergyMainFrame, "TOPLEFT", left, top)
	ComergyMovingFrame:SetPoint("BOTTOMRIGHT", ComergyMainFrame, "BOTTOMRIGHT", right, bottom)

end

function TextChanged()
	local combinedText = ""
	local text
	if (Comergy_Config.EnergyText) then
		text = combinedText .. status.curEnergy
		combinedText = text
		if (Comergy_Config.ComboText) then
			if (Comergy_Config.VerticalBars) then
				text = combinedText .. "\n"
			else
				text = combinedText .. " / "
			end
			combinedText = text
		end
	end
	if (Comergy_Config.ComboText) then
		text = combinedText .. status.curCombo
		combinedText = text .. " P"
	end
	ComergyText:SetText(combinedText)
	ComergyEnergyText:SetText(status.curEnergy)
	combinedText = status.curCombo .. " P"
	ComergyComboText:SetText(combinedText)
end

function TextStyleChanged()
	ComergyText:SetFont(getglobal(ComergyTextFonts[Comergy_Config.TextFont][2]):GetFont(), Comergy_Config.TextHeight)
	ComergyEnergyText:SetFont(getglobal(ComergyTextFonts[Comergy_Config.TextFont][2]):GetFont(), Comergy_Config.TextHeight)
	ComergyComboText:SetFont(getglobal(ComergyTextFonts[Comergy_Config.TextFont][2]):GetFont(), Comergy_Config.TextHeight)

	ComergyText:SetTextColor(Comergy_Config.TextColor[1], Comergy_Config.TextColor[2], Comergy_Config.TextColor[3])
	ComergyEnergyText:SetTextColor(Comergy_Config.TextColor[1], Comergy_Config.TextColor[2], Comergy_Config.TextColor[3])
	ComergyComboText:SetTextColor(Comergy_Config.TextColor[1], Comergy_Config.TextColor[2], Comergy_Config.TextColor[3])

	if (Comergy_Config.TextCenter) then
		ComergyText:Show()
		ComergyEnergyText:Hide()
		ComergyComboText:Hide()
	else
		ComergyText:Hide()
		if (Comergy_Config.EnergyText) then
			ComergyEnergyText:Show()
		else
			ComergyEnergyText:Hide()
		end
		if (Comergy_Config.ComboText) then
			ComergyComboText:Show()
		else
			ComergyComboText:Hide()
		end
	end

	ComergyEnergyText:SetText("100")
	ComergyEnergyText:SetWidth(ComergyEnergyText:GetStringWidth() + 5)
	ComergyComboText:SetText("0 P")
	ComergyComboText:SetWidth(ComergyComboText:GetStringWidth() + 5)
end

function Initialize()

	local defaultSettings = {
		Version = "Release 1.39",
		VersionInternal = 13900,
		ShowInCombat = false,
		ShowInStealth = true,
		ShowWhenEnergyNotFull = true,
		Locked = false,
		CritSound = true,
		StealthSound = false,
		Spacing = 4,
		Width = 164,
		ComboHeight = 5,
		EnergyHeight = 5,
		FlipBars = false,
		FlipOrientation = false,
		VerticalBars = false,

		EnergyThreshold1 = 35,
		EnergyThreshold2 = 45,
		EnergyThreshold3 = 60,
		EnergyThreshold4 = 80,
		
		EnergyColor0 = { 1, 0, 0 },
		EnergyColor1 = { 1, 0, 0 },
		EnergyColor2 = { 1, 0.5, 0 },
		EnergyColor3 = { 1, 1, 0 },
		EnergyColor4 = { 0, 1, 0 },
		EnergyColor5 = { 0, 0.625, 0.8 },

		SoundEnergy1 = false,
		SoundEnergy2 = false,
		SoundEnergy3 = true,
		SoundEnergy4 = false,
		SoundEnergy5 = false,
		SplitEnergy1 = false,
		SplitEnergy2 = false,
		SplitEnergy3 = true,
		SplitEnergy4 = false,
		EnergyText = false,
		UnifiedEnergyColor = true,

		SoundCombo1 = false,
		SoundCombo2 = false,
		SoundCombo3 = false,
		SoundCombo4 = true,
		SoundCombo5 = true,

		ComboColor0 = { 0.5, 0.5, 0.5 },
		ComboColor1 = { 1, 0, 0 },
		ComboColor2 = { 1, 0.5, 0 }, 
		ComboColor3 = { 1, 1, 0 },
		ComboColor4 = { 0, 1, 0 },
		ComboColor5 = { 0, 0.625, 0.8 },
		ComboText = false,
		ComboBGAlpha = 0.1,
		UnifiedComboColor = false,
		ComboFlash = true,

		TextColor = {1, 1, 1},
		TextHeight = 13,
		TextFont = 3,
		TextCenter = true,
		BGColor = {0, 0, 0},
		BGAlpha = 0.6,

		EnergyBGColor = {0.3, 0.3, 1},
		EnergyBGAlpha = 0.15,
		EnergyBGFlash = true,

		BarTexture = 5,
		DurationScale = 0.8,
	}

	if ((not Comergy_Config) or (not Comergy_Config.VersionInternal) or 
			(Comergy_Config.VersionInternal < BACKWARD_COMPATIBILITY)) then
		Comergy_Config = defaultSettings
	else
		if (Comergy_Config.VersionInternal ~= defaultSettings.VersionInternal) then
			Comergy_Config.VersionInternal = defaultSettings.VersionInternal
			Comergy_Config.Version = defaultSettings.Version
		end
		for i, v in pairs(defaultSettings) do
			if (Comergy_Config[i] == nil) then
				Comergy_Config[i] = v
			end
		end
	end

	if (Comergy_Config.BarTexture > #(ComergyBarTextures)) then
		Comergy_Config.BarTexture = defaultSettings.BarTexture
	end
	if (Comergy_Config.Width > 800) then
		Comergy_Config.Width = defaultSettings.Width
	end
	if (Comergy_Config.TextHeight < 6) then
		Comergy_Config.TextHeight = defaultSettings.TextHeight
	end
	if (Comergy_Config.TextFont > #(ComergyTextFonts)) then
		Comergy_Config.TextFont = defaultSettings.TextFont
	end

	status.playerInCombat = false
	status.playerInStealth = false
	status.playerEnergyEnabled = (status.playerClass == ROGUE)
		or ((status.playerClass == DRUID) and (UnitPowerType("player") == 3))

	status.curEnergy = 0
	status.maxEnergy = 100

	status.curCombo = 0

	SPELL_NAME_PROWL = GetSpellInfo(SPELL_ID_PROWL)

	for i = 1, ENERGY_SUBBAR_NUM do
		local temp = getglobal("ComergyEnergyBar"..i)
		temp:SetHeight(5)
		local initValues = { 0, 1, 1, 1 }
		InitObject(temp, initValues)

		temp.bg = temp:CreateTexture(nil, "BORDER")
		temp.bg:SetAllPoints(temp)
		initValues = { Comergy_Config.EnergyBGAlpha }
		InitObject(temp.bg, initValues)
	end
	status.energyFlashing = 0

	for i = 1, 5 do
		comboBars[i] = getglobal("ComergyComboBar"..i)
		comboBars[i]:SetMinMaxValues(0, 1)
		comboBars[i]:SetValue(1)
		comboBars[i]:SetAlpha(Comergy_Config.ComboBGAlpha)

		local initValues = { Comergy_Config.ComboBGAlpha, Comergy_Config["ComboColor"..i][1], Comergy_Config["ComboColor"..i][2], Comergy_Config["ComboColor"..i][3] }
		InitObject(comboBars[i], initValues)
	end
	status.comboFlashing = 0

	ComergyMainFrame:SetAlpha(0)
	InitObject(ComergyMainFrame, { 0 })

	SlashCmdList["COMERGY"] = function()
			if(not IsAddOnLoaded("Comergy_Options")) then
				local loaded, reason = LoadAddOn("Comergy_Options")
				if (loaded) then
					ComergyOptToggle()
				else
					DEFAULT_CHAT_FRAME:AddMessage(reason)
				end
			else
				ComergyOptToggle()
			end
		end
	SLASH_COMERGY1 = "/comergy"
	SLASH_COMERGY2 = "/cmg"

	ComergyMainFrame:Hide()

	status.initialized = true
end

-- END OF LOCAL FUNCTIONS --

function ComergyOnConfigChange()

	if (Comergy_Config.DurationScale ~= 0) then
		for i, v in pairs(DURATIONS) do
			DURATIONS[i][1] = DURATIONS[i][2] * Comergy_Config.DurationScale
		end
	else
		for i, v in pairs(DURATIONS) do
			DURATIONS[i][1] = INSTANT_DURATION
		end
	end

	if (Comergy_Config.Locked) then
		ComergyMovingFrame:EnableMouse(false)
	else
		ComergyMovingFrame:EnableMouse(true)
	end

	FrameResize()

	for i = 1, 5 do
		local color = Comergy_Config["ComboColor"..i]
		local j
		for j = 1, 3 do
			GradientObject(comboBars[i], j + 1, DURATIONS["COMBO_SHOW_DURATION"][1], color[j])
		end
	end

	ComergyBG:SetTexture(Comergy_Config.BGColor[1], Comergy_Config.BGColor[2], Comergy_Config.BGColor[3], Comergy_Config.BGAlpha)

	SplitEnergyBar()
	for i = 1, #(energyBars) do
		GradientObject(energyBars[i].bg, 1, INSTANT_DURATION, Comergy_Config.EnergyBGAlpha)
	end

	if (not Comergy_Config.UnifiedEnergyColor) then
		for i = 1, #(energyBars) do
			local color = Comergy_Config["EnergyColor"..i]
			local j
			for j = 1, 3 do
				GradientObject(energyBars[i], j + 1, DURATIONS["ENERGY_DEC_DURATION"][1], energyBars[i].maxColor[j])
			end
		end
	end
	EnergyChanged(false)
	ComboChanged()

	TextChanged()
	TextStyleChanged()

	BGResize()

	MainFrameToggle()
end

function ComergyOnLoad(self)
	status.initialized = false

	local playerClass = select(2, UnitClass("player"))

	if ((playerClass == "ROGUE") or (playerClass == "DRUID")) then
		if (playerClass == "ROGUE") then
			status.playerClass = ROGUE
		else
			status.playerClass = DRUID
		end

		self:SetScript("OnEvent", function(self, event, ...)
			EventHandlers[event](...)
		end)
		
		self:RegisterEvent("ADDON_LOADED")
		self:RegisterEvent("PLAYER_LOGIN")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_UNGHOST")
		self:RegisterEvent("PLAYER_ALIVE")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:RegisterEvent("UNIT_COMBO_POINTS")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UNIT_MAXPOWER")
		self:RegisterEvent("UPDATE_SHAPESHIFT_FORM")
		self:RegisterEvent("UPDATE_STEALTH")

	else
		ComergyMainFrame:Hide()
	end
end

-- Event Handlers --

EventHandlers = {}

function EventHandlers.ADDON_LOADED(addonName)
	if (addonName == "Comergy") then
		local COMERGY_WELCOME
		if (GetLocale() == "zhCN") then
			COMERGY_WELCOME = "Comergy已载入！使用'/cmg'打开设置窗口"
		elseif (GetLocale() == "zhTW") then
			COMERGY_WELCOME = "Comergy已載入！使用'/cmg'打開設置窗口"
		else
			COMERGY_WELCOME = "Comergy loaded. Use '/cmg' for options."
		end

		Initialize()
		ComergyOnConfigChange()
		RestorePosition()
		ComergySavePosition()
		DEFAULT_CHAT_FRAME:AddMessage(COMERGY_WELCOME)
		status.initialized = true
	end
end

function EventHandlers.PLAYER_LOGIN()
	playerGUID = UnitGUID("player")
end

-- For Druid cat form change and rogue stealth
function EventHandlers.UPDATE_SHAPESHIFT_FORM()
	if (not status.initialized) then
		return
	end

	if (status.playerClass == DRUID) then
		status.playerEnergyEnabled = (UnitPowerType("player") == 3)
		if (status.playerEnergyEnabled) then
			MaxEnergyChanged()
		end
		MainFrameToggle()
	end

	-- Shadow dance
	if (status.playerClass == ROGUE) then
		local form = GetShapeshiftForm()
		if ((form == 3) and (Comergy_Config.StealthSound)) then		
			PlaySoundFile("Sound\\interface\\iQuestUpdate.wav")
		end
		status.playerInStealth = ((form > 0) and (form < 4))
		MainFrameToggle()
	end
end

function EventHandlers.UPDATE_STEALTH()
	-- Stealth
	if (status.playerClass == ROGUE) then
		local form = GetShapeshiftForm()
		if ((form > 0) and (form < 3) and (Comergy_Config.StealthSound)) then
			PlaySoundFile("Sound\\interface\\iQuestUpdate.wav")
		end
		status.playerInStealth = ((form > 0) and (form < 4))
		MainFrameToggle()
	end
end

function EventHandlers.COMBAT_LOG_EVENT_UNFILTERED(...)
	if (not status.initialized) then
		return
	end
	
	if (select(3, ...) == UnitGUID("player")) then
		local type = select(2, ...)
		if ((type == "SPELL_DAMAGE") and (Comergy_Config.CritSound)) then
			if (select(18, ...)) then
				PlaySoundFile("Interface\\AddOns\\Comergy\\Sound\\critical.mp3")
			end
		elseif ((type == "SPELL_CAST_FAILED") and (select(12, ...) == ERR_OUT_OF_ENERGY)
				and (Comergy_Config.EnergyBGFlash)) then
			if (status.energyFlashing == 0) then
				status.energyFlashing = 2
				for i = 1, #(energyBars) do
					ResetObject(energyBars[i].bg, FLASH_VALUES)
					energyBars[i].bg:SetTexture(Comergy_Config.EnergyBGColor[1], Comergy_Config.EnergyBGColor[2],
									Comergy_Config.EnergyBGColor[3], FLASH_HIGH)
					GradientObject(energyBars[i].bg, 1, FLASH_DURATION, Comergy_Config.EnergyBGAlpha)
				end
			end
		elseif ((type == "SPELL_AURA_APPLIED") or (type == "SPELL_AURA_REFRESH")) then
			if (select(10, ...) == SPELL_NAME_PROWL) then
				status.playerInStealth = true
				MainFrameToggle()
			end
		elseif (type == "SPELL_AURA_REMOVED") then
			if (select(10, ...) == SPELL_NAME_PROWL) then
				status.playerInStealth = false
				MainFrameToggle()
			end
		end
	end
end

function EventHandlers.PLAYER_REGEN_DISABLED()
	if (not status.initialized) then
		return
	end

	status.playerInCombat = true
	MainFrameToggle()
end

function EventHandlers.PLAYER_REGEN_ENABLED()
	if (not status.initialized) then
		return
	end

	status.playerInCombat = false
	MainFrameToggle()
end

function ReEnterWorld()
	EventHandlers.UPDATE_SHAPESHIFT_FORM()
	MaxEnergyChanged()
	ComboChanged()
end

function EventHandlers.PLAYER_ALIVE()
	if (not status.initialized) then
		return
	end

	ReEnterWorld()
end

function EventHandlers.PLAYER_UNGHOST()
	if (not status.initialized) then
		return
	end

	ReEnterWorld()
end

function EventHandlers.PLAYER_ENTERING_WORLD()
	if (not status.initialized) then
		return
	end

	ReEnterWorld()
end

function EventHandlers.UNIT_COMBO_POINTS()
	if (not status.initialized) then
		return
	end
	if (not status.playerEnergyEnabled) then
		return
	end

	ComboChanged()
end

function EventHandlers.PLAYER_TARGET_CHANGED()
	if (not status.initialized) then
		return
	end
	if (not status.playerEnergyEnabled) then
		return
	end

	ComboChanged()
end

function EventHandlers.UNIT_MAXPOWER()
	if (not status.initialized) then
		return
	end

	MaxEnergyChanged()
end

function ComergySavePosition()
	local parent = UIParent
	local frame = ComergyMainFrame

	local s = frame:GetScale()
	local left, top = frame:GetLeft() * s, frame:GetTop() * s
	local right, bottom = frame:GetRight() * s, frame:GetBottom() * s
	local pwidth, pheight = parent:GetWidth(), parent:GetHeight()

	local x, y, point
	if left < (pwidth-right) and left < abs((left+right)/2 - pwidth/2) then
		x = left
		point="LEFT"
	elseif (pwidth-right) < abs((left+right)/2 - pwidth/2) then
		x = right-pwidth
		point="RIGHT"
	else
		x = (left+right)/2 - pwidth/2
		point=""
	end
	
	if bottom < (pheight-top) and bottom < abs((bottom+top)/2 - pheight/2) then
		y = bottom
		point="BOTTOM"..point
	elseif (pheight-top) < abs((bottom+top)/2 - pheight/2) then
		y = top-pheight
		point="TOP"..point
	else
		y = (bottom+top)/2 - pheight/2
	end
	
	if point=="" then
		point = "CENTER"
	end
	
	Comergy_Config.X = x
	Comergy_Config.Y = y
	Comergy_Config.Point = point
end


function RestorePosition()
	local parent = UIParent
	local frame = ComergyMainFrame
	
	if ((not Comergy_Config.X) or (not Comergy_Config.Y) or (not Comergy_Config.Point)) then
		Comergy_Config.X = 0
		Comergy_Config.Y = 0
		Comergy_Config.Point = "CENTER"
	end

	local x = Comergy_Config.X
	local y = Comergy_Config.Y
	local point = Comergy_Config.Point
	local s = ComergyMainFrame:GetScale()

	x = x / s
	y = y / s
	
	frame:ClearAllPoints()
	frame:SetPoint(point, parent, point, x, y)
end

function ComergyOnUpdate(self, elapsed)
	if (not status.initialized) then
		return
	end

	for i = 1, 5 do
		UpdateObject(comboBars[i], elapsed)
		comboBars[i]:SetAlpha(comboBars[i].curValue[1])
		comboBars[i]:SetStatusBarColor(comboBars[i].curValue[2], comboBars[i].curValue[3], comboBars[i].curValue[4])
	end

	local curEnergy = UnitMana("player")
	if (status.curEnergy ~= curEnergy) then
		local inc = (curEnergy > status.curEnergy)
		status.curEnergy = curEnergy
		EnergyChanged(inc)
	end

	for i = 1, #(energyBars) do
		UpdateObject(energyBars[i], elapsed)
		SetStatusBarValue(energyBars[i], energyBars[i].curValue[1])
		energyBars[i]:SetStatusBarColor(energyBars[i].curValue[2], energyBars[i].curValue[3], energyBars[i].curValue[4])

		UpdateObject(energyBars[i].bg, elapsed)
		energyBars[i].bg:SetTexture(Comergy_Config.EnergyBGColor[1], Comergy_Config.EnergyBGColor[2],
								Comergy_Config.EnergyBGColor[3], energyBars[i].bg.curValue[1])
	end
	if ((status.energyFlashing > 0) and (energyBars[1].bg.curValue[1] == Comergy_Config.EnergyBGAlpha)) then
		status.energyFlashing = status.energyFlashing - 1
		if (status.energyFlashing > 0) then
			for i = 1, #(energyBars) do
				ResetObject(energyBars[i].bg, FLASH_VALUES)
				GradientObject(energyBars[i].bg, 1, FLASH_DURATION, Comergy_Config.EnergyBGAlpha)
			end
		end
	end

	if ((status.comboFlashing > 0) and (comboBars[1].curValue[1] == Comergy_Config.ComboBGAlpha)) then
		status.comboFlashing = status.comboFlashing - 1
		if (status.comboFlashing > 0) then
			for i = 1, #(comboBars) do
				ResetObject(comboBars[i], FLASH_VALUES)
				GradientObject(comboBars[i], 1, FLASH_DURATION, Comergy_Config.ComboBGAlpha)
			end
		else
			for i = 1, #(comboBars) do
				ResetObject(comboBars[i], ONE_VALUES)
			end
		end
	end

	UpdateObject(ComergyMainFrame, elapsed)
	ComergyMainFrame:SetAlpha(ComergyMainFrame.curValue[1])
	if (ComergyMainFrame.curValue[1] == 0) then
		ComergyMainFrame:Hide()
	end
end