local sliders = {}
local checkButtons = {}

local function ChooseTexture(i)
	Comergy_Config.BarTexture = i
	ComergyTextureDropdownText:SetText(ComergyBarTextures[i][1])
	ComergyOnConfigChange()
end

local function ChooseFont(i)
	Comergy_Config.TextFont = i
	ComergyTextDropdownText:SetText(ComergyTextFonts[i][1])
	ComergyOnConfigChange()
end

local function CreateSlider(name, min, max, parent, x, y, width, valueStep, title, minText, maxText)
	local slider = CreateFrame("Slider", "ComergyOptSlider"..name, parent, "ComergyOptSliderTemplate")
	slider:SetMinMaxValues(min, max)
	slider:SetWidth(width)
	slider:SetHeight(16)
	slider:SetPoint("TOPLEFT", x, y)
	slider:SetValueStep(valueStep)
	getglobal(slider:GetName().."Text"):SetText(title)
	if (minText) then
		getglobal(slider:GetName().."Low"):SetText(minText)
		getglobal(slider:GetName().."High"):SetText(maxText)
	else
		getglobal(slider:GetName().."Low"):SetText(min)
		getglobal(slider:GetName().."High"):SetText(max)
	end


	table.insert(sliders, slider)
end

local function CreateCheckButton(name, parent, x, y)
	local checkButton = CreateFrame("CheckButton", "ComergyOptCheckButton"..name, parent, "ComergyOptCheckButtonTemplate")
	checkButton:SetPoint("TOPLEFT", x, y)
	if (CHECKOPTINFO[name]) then
		getglobal(checkButton:GetName().."Text"):SetTextColor(1, 1, 1)
		getglobal(checkButton:GetName().."Text"):SetText(CHECKOPTINFO[name])
	end

	table.insert(checkButtons, checkButton)
end

local function CreateEditBox(name, parent, x, y)
	local editBox = CreateFrame("EditBox", "ComergyOptEditBox"..name, parent, "ComergyOptEditBoxTemplate")
	editBox:SetPoint("TOPLEFT", x, y)
	return editBox
end

local function QuitColor(issave, name)
	local r, g, b
	if issave then
		r, g, b = ColorPickerFrame:GetColorRGB()
	else
		local c = ColorPickerFrame.previousValues
		r, g, b = c[1], c[2], c[3]
	end
	Comergy_Config[name][1] = r
	Comergy_Config[name][2] = g
	Comergy_Config[name][3] = b
	ColorPickerFrame.button:GetNormalTexture():SetVertexColor(r, g, b)
	ComergyOnConfigChange()

end

local function CreateColorButton(name, parent, x, y)
	local colorButton = CreateFrame("Button", name, parent)
	colorButton:SetNormalTexture("Interface/ChatFrame/ChatFrameColorSwatch")

	local bg = colorButton:CreateTexture(nil, "BACKGROUND")
	bg:SetWidth(14)
	bg:SetHeight(14)
	bg:SetTexture(0, 0, 0)
	bg:SetPoint("CENTER")
	
	local label = colorButton:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	label:SetPoint("RIGHT", colorButton, "LEFT")
	if (COLORPICKERINFO[name]) then
		label:SetText(COLORPICKERINFO[name])
	end
	colorButton:SetWidth(24)
	colorButton:SetHeight(24)
	
	colorButton:RegisterForClicks("LeftButtonUp")
	colorButton:SetScript("OnClick", function(self)
		ComergyOptClearEditBoxFocus()
		if ColorPickerFrame:IsShown() then
			ColorPickerFrame:Hide()
		else
			local color = Comergy_Config[name]
			ColorPickerFrame.button = self
			ColorPickerFrame.previousValues = {color[1], color[2], color[3]}
			
			ColorPickerFrame.func = function() QuitColor(true, name) end
			ColorPickerFrame.cancelFunc = function() QuitColor(false, name) end
			
			ColorPickerFrame:SetColorRGB(color[1], color[2], color[3])
			
			ShowUIPanel(ColorPickerFrame)
		end
	end)
	colorButton:SetScript("OnShow", function(self)
		local color = Comergy_Config[name]
		self:GetNormalTexture():SetVertexColor(color[1], color[2], color[3])
	end)

	colorButton:SetPoint("TOPLEFT", parent, "TOPLEFT", x, y)
	colorButton:Show()	
	return colorButton
end

local function DisableCheckBox(checkBox)
	checkBox:Disable()
	checkBox:SetAlpha(0.5)
end

local function EnableCheckBox(checkBox)
	checkBox:Enable()
	checkBox:SetAlpha(1)
end

local function SetCheckBox()
	if (ComergyOptCheckButtonShowInCombat:GetChecked()) then
		EnableCheckBox(ComergyOptCheckButtonShowWhenEnergyNotFull)
		EnableCheckBox(ComergyOptCheckButtonShowInStealth)
	else
		DisableCheckBox(ComergyOptCheckButtonShowWhenEnergyNotFull)
		DisableCheckBox(ComergyOptCheckButtonShowInStealth)
	end
end

local function Check()
	local i
	for i = 1, #(checkButtons) do
		local _, _, name = string.find(checkButtons[i]:GetName(), "ComergyOptCheckButton(.+)")
		checkButtons[i]:SetChecked(Comergy_Config[name])
	end
	SetCheckBox()
end

local function Slide()
	local i
	for i = 1, #(sliders) do
		local _, _, name = string.find(sliders[i]:GetName(), "ComergyOptSlider(.+)")
		sliders[i]:SetValue(Comergy_Config[name])
	end
end

function ComergyOpt_OnLoad()
	--Options Frame needs to be UISpecialFrame
	table.insert(UISpecialFrames,"ComergyOptFrame")

	CreateSlider("Width", 50, 800, ComergyOptGeneralFrame, 25, -135, 250, 1, COMERGY_WIDTH)
	CreateSlider("Spacing", 0, 5, ComergyOptGeneralFrame, 25, -170, 110, 1, COMERGY_SPACING)
	CreateSlider("BGAlpha", 0, 1, ComergyOptGeneralFrame, 160, -170, 110, 0.05, COMERGY_BG_ALPHA)
	CreateSlider("DurationScale", 0, 2, ComergyOptGeneralFrame, 25, -205, 110, 0.2, COMERGY_DURATION_SCALE, COMERGY_ACCURATE, COMERGY_FADE)
	CreateSlider("TextHeight", 6, 26, ComergyOptGeneralFrame, 160, -205, 110, 1, COMERGY_FONT_SIZE)

	CreateSlider("EnergyHeight", 0, 16, ComergyOptEnergyFrame, 25, -230, 110, 1, COMERGY_ENERGY_HEIGHT)
	CreateSlider("EnergyBGAlpha", 0, 0.3, ComergyOptEnergyFrame, 160, -230, 110, 0.02, COMERGY_ENERGY_BG_ALPHA)

	CreateSlider("ComboBGAlpha", 0, 0.3, ComergyOptComboFrame, 25, -230, 110, 0.02, COMERGY_COMBO_BG_ALPHA)
	CreateSlider("ComboHeight", 0, 16, ComergyOptComboFrame, 160, -230, 110, 1, COMERGY_COMBO_HEIGHT)

	CreateCheckButton("ShowInCombat", ComergyOptGeneralFrame, 13, -10)
	ComergyOptCheckButtonShowInCombat:SetScript("OnClick", function() 
			ComergyOptCheckButton_OnClick(ComergyOptCheckButtonShowInCombat)
			SetCheckBox()
		end)
	CreateCheckButton("ShowWhenEnergyNotFull", ComergyOptGeneralFrame, 43, -30)
	CreateCheckButton("ShowInStealth", ComergyOptGeneralFrame, 43, -50)
	CreateCheckButton("Locked", ComergyOptGeneralFrame, 13, -80)
	CreateCheckButton("CritSound", ComergyOptGeneralFrame, 163, -80)
	CreateCheckButton("StealthSound", ComergyOptGeneralFrame, 13, -100)
	CreateCheckButton("FlipBars", ComergyOptGeneralFrame, 13, -290)
	CreateCheckButton("FlipOrientation", ComergyOptGeneralFrame, 13, -310)
	CreateCheckButton("TextCenter", ComergyOptGeneralFrame, 163, -290)
	CreateCheckButton("VerticalBars", ComergyOptGeneralFrame, 163, -310)

	CreateCheckButton("SoundEnergy1", ComergyOptEnergyFrame, 120, -60)
	CreateCheckButton("SoundEnergy2", ComergyOptEnergyFrame, 120, -90)
	CreateCheckButton("SoundEnergy3", ComergyOptEnergyFrame, 120, -120)
	CreateCheckButton("SoundEnergy4", ComergyOptEnergyFrame, 120, -150)
	CreateCheckButton("SoundEnergy5", ComergyOptEnergyFrame, 120, -180)
	CreateCheckButton("SplitEnergy1", ComergyOptEnergyFrame, 170, -60)
	CreateCheckButton("SplitEnergy2", ComergyOptEnergyFrame, 170, -90)
	CreateCheckButton("SplitEnergy3", ComergyOptEnergyFrame, 170, -120)
	CreateCheckButton("SplitEnergy4", ComergyOptEnergyFrame, 170, -150)

	CreateCheckButton("EnergyText", ComergyOptEnergyFrame, 20, -260)
	CreateCheckButton("UnifiedEnergyColor", ComergyOptEnergyFrame, 20, -280)
	CreateCheckButton("EnergyBGFlash", ComergyOptEnergyFrame, 20, -300)

	CreateCheckButton("SoundCombo1", ComergyOptComboFrame, 80, -69)
	CreateCheckButton("SoundCombo2", ComergyOptComboFrame, 80, -99)
	CreateCheckButton("SoundCombo3", ComergyOptComboFrame, 80, -129)
	CreateCheckButton("SoundCombo4", ComergyOptComboFrame, 80, -159)
	CreateCheckButton("SoundCombo5", ComergyOptComboFrame, 80, -189)
	CreateCheckButton("ComboText", ComergyOptComboFrame, 20, -260)
	CreateCheckButton("UnifiedComboColor", ComergyOptComboFrame, 20, -280)
	CreateCheckButton("ComboFlash", ComergyOptComboFrame, 20, -300)

	CreateEditBox("EnergyThreshold1", ComergyOptEnergyFrame, 30, -53):SetID(1)
	CreateEditBox("EnergyThreshold2", ComergyOptEnergyFrame, 30, -83):SetID(2)
	CreateEditBox("EnergyThreshold3", ComergyOptEnergyFrame, 30, -113):SetID(3)
	CreateEditBox("EnergyThreshold4", ComergyOptEnergyFrame, 30, -143):SetID(4)
	
	CreateColorButton("BGColor", ComergyOptGeneralFrame, 230, -230)
	CreateColorButton("TextColor", ComergyOptGeneralFrame, 230, -260)

	CreateColorButton("EnergyColor0", ComergyOptEnergyFrame, 80, -30)
	CreateColorButton("EnergyColor1", ComergyOptEnergyFrame, 80, -60)
	CreateColorButton("EnergyColor2", ComergyOptEnergyFrame, 80, -90)
	CreateColorButton("EnergyColor3", ComergyOptEnergyFrame, 80, -120)
	CreateColorButton("EnergyColor4", ComergyOptEnergyFrame, 80, -150)
	CreateColorButton("EnergyColor5", ComergyOptEnergyFrame, 80, -180)

	CreateColorButton("EnergyBGColor", ComergyOptEnergyFrame, 250, -260)

	CreateColorButton("ComboColor0", ComergyOptComboFrame, 45, -39)
	CreateColorButton("ComboColor1", ComergyOptComboFrame, 45, -69)
	CreateColorButton("ComboColor2", ComergyOptComboFrame, 45, -99)
	CreateColorButton("ComboColor3", ComergyOptComboFrame, 45, -129)
	CreateColorButton("ComboColor4", ComergyOptComboFrame, 45, -159)
	CreateColorButton("ComboColor5", ComergyOptComboFrame, 45, -189)

	ComergyOptTitle:SetText("Comergy "..Comergy_Config.Version)
	ComergyOptTab1:SetText(COMERGY_GENERAL)
	ComergyOptTab2:SetText(COMERGY_ENERGY)
	ComergyOptTab3:SetText(COMERGY_COMBO)
	ComergyOptTab1:Show()
	ComergyOptTab2:Show()
	ComergyOptTab3:Show()

	ComergyOptComboTextSound:SetText(COMERGY_COMBO_TEXT_SOUND)
	ComergyOptEnergyTextSound:SetText(COMERGY_ENERGY_TEXT_SOUND)
	ComergyOptEnergyTextSplit:SetText(COMERGY_ENERGY_TEXT_SPLIT)

	local tdd = CreateFrame("Frame", "ComergyTextureDropdown", ComergyOptGeneralFrame, "UIDropDownMenuTemplate")
	ComergyTextureDropdownMiddle:SetWidth(50)
	tdd:SetPoint("TOPLEFT", 55, -230)
	ComergyTextureDropdownText:SetText(ComergyBarTextures[Comergy_Config.BarTexture][1])

	ComergyTextureDropdownButton:SetScript("OnClick", function()
		if (DropDownList1:IsShown()) then
			DropDownList1:Hide()
		else
			local texturesList = {}
			for i = 1, #(ComergyBarTextures) do
				table.insert(texturesList,
					{ text = ComergyBarTextures[i][1], func = function() ChooseTexture(i) end, checked = false })
			end
			texturesList[Comergy_Config.BarTexture].checked = true
			EasyMenu(texturesList, ComergyTextureDropdown, ComergyTextureDropdown, 0 , 0, nil)
			ToggleDropDownMenu(1, nil, ComergyTextureDropdown, ComergyTextureDropdown, 0, 0, texturesList)
			DropDownList1:Show()
		end
	end)
	ComergyOptGeneralTextTexture:SetJustifyH("RIGHT")
	ComergyOptGeneralTextTexture:SetPoint("BOTTOMRIGHT", ComergyTextureDropdownButton, "BOTTOMLEFT", -45, 0)
	ComergyOptGeneralTextTexture:SetPoint("TOPLEFT", ComergyTextureDropdownButton, "TOPLEFT", -125, 0)
	ComergyOptGeneralTextTexture:SetText(COMERGY_TEXTURE)

	local tdd2 = CreateFrame("Frame", "ComergyTextDropdown", ComergyOptGeneralFrame, "UIDropDownMenuTemplate")
	ComergyTextDropdownMiddle:SetWidth(70)
	tdd2:SetPoint("TOPLEFT", 35, -260)
	ComergyTextDropdownText:SetText(ComergyTextFonts[Comergy_Config.TextFont][1])

	ComergyTextDropdownButton:SetScript("OnClick", function()
		if (DropDownList1:IsShown()) then
			DropDownList1:Hide()
		else
			local fontsList = {}
			for i = 1, #(ComergyTextFonts) do
				table.insert(fontsList,
					{ text = ComergyTextFonts[i][1], func = function() ChooseFont(i) end, checked = false })
			end
			fontsList[Comergy_Config.TextFont].checked = true
			EasyMenu(fontsList, ComergyTextDropdown, ComergyTextDropdown, 0 , 0, nil)
			ToggleDropDownMenu(1, nil, ComergyTextDropdown, ComergyTextDropdown, 0, 0, fontsList)
			DropDownList1:Show()
		end
	end)
	ComergyOptGeneralTextText:SetJustifyH("RIGHT")
	ComergyOptGeneralTextText:SetPoint("BOTTOMRIGHT", ComergyTextDropdownButton, "BOTTOMLEFT", -65, 0)
	ComergyOptGeneralTextText:SetPoint("TOPLEFT", ComergyTextDropdownButton, "TOPLEFT", -145, 0)
	ComergyOptGeneralTextText:SetText(COMERGY_FONT)

	Check()
	Slide()

	ComergyOptEditBoxEnergyThreshold1:SetText(Comergy_Config.EnergyThreshold1)
	ComergyOptEditBoxEnergyThreshold2:SetText(Comergy_Config.EnergyThreshold2)
	ComergyOptEditBoxEnergyThreshold3:SetText(Comergy_Config.EnergyThreshold3)
	ComergyOptEditBoxEnergyThreshold4:SetText(Comergy_Config.EnergyThreshold4)

	ComergyOptTabOnClick(1)

	ComergyOptFrame:Hide()
end

function ComergyOptCheckButton_OnClick(button)
	local _, _, name = string.find(button:GetName(),"ComergyOptCheckButton(.+)")
	
	Comergy_Config[name] = (button:GetChecked() and true or false)
	PlaySound(button:GetChecked() and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
	ComergyOnConfigChange()
end

function ComergyOptToggle()
	if (ComergyOptFrame:IsShown()) then
		ComergyOptFrame:Hide()
	else
		ComergyOptFrame:Show()
	end
end

function ComergyOptTabOnClick(id)
	PlaySound("GAMEGENERICBUTTONPRESS")
	local tab
	local i
	for i = 1, 3 do
		tab = getglobal("ComergyOptTab"..i)
		if (tab) then
			tab:UnlockHighlight()
		end
	end
	getglobal("ComergyOptTab"..id):LockHighlight()
	ComergyOptGeneralFrame:Hide()
	ComergyOptEnergyFrame:Hide()
	ComergyOptComboFrame:Hide()
	if (id == 1) then
		ComergyOptGeneralFrame:Show()
	elseif (id == 2) then
		ComergyOptEnergyFrame:Show()
	elseif (id == 3) then
		ComergyOptComboFrame:Show()
	end
end

function ComergyOptSlider_OnValueChanged(self)
	local _, _, name = string.find(self:GetName(),"ComergyOptSlider(.+)")
	Comergy_Config[name] = self:GetValue()
	ComergyOnConfigChange()
end

function ComergyOptEditBox_OnTextChanged(self)
--	self:ClearFocus()
--	local energyThreshold = { 0, Comergy_Config.EnergyThreshold1, Comergy_Config.EnergyThreshold2, Comergy_Config.EnergyThreshold3, UnitManaMax("player") }

	if (self:GetNumber() <= 0 or self:GetNumber() >= UnitManaMax("player")) then
		message(COMERGY_OUT_OF_RANGE)
		self:SetNumber(Comergy_Config["EnergyThreshold"..self:GetID()])
	else
		Comergy_Config["EnergyThreshold"..self:GetID()] = self:GetNumber()
		ComergyOnConfigChange()
	end
end

function ComergyOptEditBox_OnTextNotChanged(self)
--	self:ClearFocus()
	self:SetNumber(Comergy_Config["EnergyThreshold"..self:GetID()])
end

function ComergyOptClearEditBoxFocus()
	ComergyOptEditBoxEnergyThreshold1:ClearFocus()
	ComergyOptEditBoxEnergyThreshold2:ClearFocus()
	ComergyOptEditBoxEnergyThreshold3:ClearFocus()
end